/* filter - postprocessing of flex output through filters */

/*  This file is part of flex. */

/*  Redistribution and use in source and binary forms, with or without */
/*  modification, are permitted provided that the following conditions */
/*  are met: */

/*  1. Redistributions of source code must retain the above copyright */
/*     notice, this list of conditions and the following disclaimer. */
/*  2. Redistributions in binary form must reproduce the above copyright */
/*     notice, this list of conditions and the following disclaimer in the */
/*     documentation and/or other materials provided with the distribution. */

/*  Neither the name of the University nor the names of its contributors */
/*  may be used to endorse or promote products derived from this software */
/*  without specific prior written permission. */

/*  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR */
/*  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED */
/*  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR */
/*  PURPOSE. */

#include "flexdef.h"
static const char * check_4_gnu_m4 =
    "m4_dnl ifdef(`__gnu__', ,"
    "`errprint(Flex requires GNU M4. Set the PATH or set the M4 environment variable to its path name.)"
    " m4exit(2)')\n";

bool h_out = false;
extern FILE *flout;
extern char *flo; 
/** global chain. */
struct filter *output_chain = NULL;

/* Allocate and initialize an external filter.
 * @param chain the current chain or NULL for new chain
 * @param cmd the command to execute.
 * @param ... a NULL terminated list of (const char*) arguments to command,
 *            not including argv[0].
 * @return newest filter in chain
 */
struct filter *filter_create_ext (struct filter *chain, const char *cmd,
				  ...)
{
	struct filter *f;
	int     max_args;
	const char *s;
	va_list ap;

	/* allocate and initialize new filter */
	f = calloc(sizeof(struct filter), 1);
	if (!f)
		flexerror(_("calloc failed (f) in filter_create_ext"));
	f->filter_func = NULL;
	f->extra = NULL;
	f->next = NULL;
	f->argc = 0;

	if (chain != NULL) {
		/* append f to end of chain */
		while (chain->next)
			chain = chain->next;
		chain->next = f;
	}


	/* allocate argv, and populate it with the argument list. */
	max_args = 11;
	f->argv = malloc(sizeof(char *) * (size_t) (max_args + 1));
	if (!f->argv)
		flexerror(_("malloc failed (f->argv) in filter_create_ext"));
	f->argv[f->argc++] = cmd;

	va_start (ap, cmd);
	while ((s = va_arg (ap, const char *)) != NULL) {
		if (f->argc >= max_args) {
			max_args += 8;
			f->argv = realloc(f->argv, sizeof(char*) * (size_t) (max_args + 1));
		}
		f->argv[f->argc++] = s;
	}
	f->argv[f->argc] = NULL;

	va_end (ap);
	return f;
}

/* Allocate and initialize an internal filter.
 * @param chain the current chain or NULL for new chain
 * @param filter_func The function that will perform the filtering.
 *        filter_func should return 0 if successful, and -1
 *        if an error occurs -- or it can simply exit().
 * @param extra optional user-defined data to pass to the filter.
 * @return newest filter in chain
 */
struct filter *filter_create_int (struct filter *chain,
				  int (*filter_func) (struct filter *),
				  void *extra)
{
	struct filter *f;

	/* allocate and initialize new filter */
	f = calloc(sizeof(struct filter), 1);
	if (!f)
		flexerror(_("calloc failed in filter_create_int"));
	f->next = NULL;
	f->argc = 0;
	f->argv = NULL;

	f->filter_func = filter_func;
	f->extra = extra;

	if (chain != NULL) {
		/* append f to end of chain */
		while (chain->next)
			chain = chain->next;
		chain->next = f;
	}

	return f;
}

/** Fork and exec entire filter chain.
 *  @param chain The head of the chain.
 *  @return true on success.
 */
bool filter_apply_chain (struct filter * chain)
{
	pid_t pid;
	
	/* Tricky recursion, since we want to begin the chain
	 * at the END. Why? Because we need all the forked processes
	 * to be children of the main flex process.
	 */
	if (chain)
		filter_apply_chain (chain->next);
	else
		return true;

	if ((pid = fork ()) == -1)
		flexerror (_("fork failed"));

	if (pid == 0) {
		/* child */

		/* run as a filter, either internally or by exec */
		if (chain->filter_func) {
			int     r;

			if ((r = chain->filter_func (chain)) == -1)
				flexfatal (_("filter_func failed"));
			FLEX_EXIT (0);
		}
		else {
			execvp (chain->argv[0],
				(char **const) (chain->argv));
            lerr_fatal ( _("exec of %s failed"),
                    chain->argv[0]);
		}

		FLEX_EXIT (1);
	}

	/* Parent */
  int child_status;
    while (wait(&child_status) > 0){
       if (!WIFEXITED (child_status) || WEXITSTATUS (child_status) != 0)
         lerr_fatal (_("flex: child error from m4"));     
		}
	return true;
}

/** Truncate the chain to max_len number of filters.
 * @param chain the current chain.
 * @param max_len the maximum length of the chain.
 * @return the resulting length of the chain.
 */
int filter_truncate (struct filter *chain, int max_len)
{
	int     len = 1;

	if (!chain)
		return 0;

	while (chain->next && len < max_len) {
		chain = chain->next;
		++len;
	}

	chain->next = NULL;
	return len;
}

/** RISC OS change here, instead of teeing
 *  for a header file, write the start of two files
 *  The default is for just the c file output.
 *  @return 0 (zero) on success, and -1 on failure.
 */
int filter_tee_header (struct filter *chain)
{ 
	/* c file opened and closed in main.c  */
	if( fseek(flout,0,SEEK_CUR) == 0){
        fputs (check_4_gnu_m4, flout);
	fputs ("m4_changecom`'m4_dnl\n", flout);
	fputs ("m4_changequote`'m4_dnl\n", flout);
	fputs ("m4_changequote([[,]])[[]]m4_dnl\n", flout);
	fputs ("m4_define([[M4_YY_NOOP]])[[]]m4_dnl\n", flout);
	/* overwriting for a headerfile begins here */
	fprintf (flout, "m4_define( [[M4_YY_OUTFILE_NAME]],[[%s]])m4_dnl\n",
 		 outfilename ? outfilename : "<stdout>");
 	/* create extra space to add extra headerfile lines later */  	 
 		      if (headerfilename) {
	fprintf (flout,"m4_dnl reserve some space for %s          \n", headerfilename);
	fputs ("m4_dnl prints no output to c file                          \n", flout);
				}	 
			}
		/* file gas been closed after c file use */
	else if (headerfilename && !h_out) {
			flout = fopen (flo, "r+");
			if (flout == NULL)
		    lerr (_("hdr: error re-opening temp file %s"),flo);
		    /* overwrite temp file beginning at pos 248 */
		    if(fseek (flout,248,SEEK_SET) !=0)
	           lerr (_("error with initial seek_set in header temp %ld\n"));
				
		fputs ("m4_define( [[M4_YY_IN_HEADER]],[[]])m4_dnl\n", flout);
		fprintf (flout, "#ifndef %sHEADER_H\n", prefix);
		fprintf (flout, "#define %sHEADER_H 1\n", prefix);
		fprintf (flout, "#define %sIN_HEADER 1\n\n", prefix);
		fprintf (flout, "m4_define( [[M4_YY_OUTFILE_NAME]],[[%s]])m4_dnl\n",
 			 headerfilename ? headerfilename : "<stdout>");
 			 	if (fseek (flout,0,SEEK_END) !=0)
	                lerr (_("error with seek_end of header temp %S"),flout);     
		    	/* append header file ending */
		fprintf (flout, "\n#line 4000 \"M4_YY_OUTFILE_NAME\"\n");
		fprintf (flout, "#undef %sIN_HEADER\n", prefix);
		fprintf (flout, "#endif /* %sHEADER_H */\n", prefix);
		fputs ("m4_undefine( [[M4_YY_IN_HEADER]])m4_dnl\n", flout);

	    if (fclose (flout) != 0)
		    lerr(_("error closing h temp file %s"),flo);
		
		}
	return 0;
}

static bool is_blank_line (const char *str)
{
	while (isspace(*str))
		str++;
	return (*str == '\0');
}

/** Adjust the line numbers in the #line directives of the generated scanner.
 * After the m4 expansion, the line numbers are incorrect since the m4 macros
 * can add or remove lines.  This only adjusts line numbers for generated code,
 * not user code. This also happens to be a good place to squeeze multiple
 * blank lines into a single blank line.
 */
int filter_fix_linedirs (struct filter *chain)
{
	char    *buf = malloc(512);
	const size_t readsz = 512;
	int     lineno = 1;
	bool    in_gen = true;	/* in generated code */
	bool    last_was_blank = false;

    extern char *m4o;
    
	if (!chain)
		return 0;
		
	 FILE *flfn;

	if (headerfilename && h_out==true) 
	outfilename = headerfilename;

	if (use_stdout)
	flfn = stdout;
	else {
		if (outfilename) {
		remove (outfilename);
	flfn = fopen (outfilename, "w");
	if (flfn == NULL)
	lerr (_("flex out: error opening output file %s"),
 			outfilename );
 			setlinebuf(flfn);
		}
	}
	
	FILE *m4fn = fopen (m4o, "r");
	     if (m4fn == NULL)
	     lerr (_("flex out: error opening m4 file %s"),m4o);
		
	while (fgets (buf, (int) readsz, m4fn)) {

		regmatch_t m[10];

		/* Check for #line directive. */
		if (buf[0] == '#'
			&& regexec (&regex_linedir, buf, 3, m, 0) == 0) {

			char   *fname;

			/* extract the line number and filename */
			fname = regmatch_dup (&m[2], buf);

			if (strcmp (fname,
				outfilename ? outfilename : "<stdout>")
					== 0
			 || strcmp (fname,
			 	headerfilename ? headerfilename : "<stdout>")
					== 0) {

				char    *s1, *s2;
				char	filename[MAXLINE];

				s1 = fname;
				s2 = filename;

				while ((s2 - filename) < (MAXLINE - 1) && *s1) {
					/* Escape the backslash */
					if (*s1 == '\\')
						*s2++ = '\\';
					/* Escape the double quote */
					if (*s1 == '\"')
						*s2++ = '\\';
					/* Copy the character as usual */
					*s2++ = *s1++;
				}

				*s2 = '\0';

				/* Adjust the line directives. */
				in_gen = true;
				snprintf (buf, readsz, "#line %d \"%s\"\n",
					  lineno + 1, filename);
			}
			else {
				/* it's a #line directive for code we didn't write */
				in_gen = false;
			}

			free (fname);
			last_was_blank = false;
		}

		/* squeeze blank lines from generated code */
		else if (in_gen && is_blank_line(buf)) {
			if (last_was_blank)
				continue;
			else
				last_was_blank = true;
		}

		else {
			/* it's a line of normal, non-empty code. */
			last_was_blank = false;
		}

		fputs (buf, flfn);
		lineno++;
	}
            free(buf);
 		if (fclose (m4fn) != 0)
		    lerr (_("flex out: error closing m4 file %s"),m4o);
  		if (fclose (flfn) != 0)
		    lerr (_("flex out: error closing output %s"),
		    outfilename ? outfilename : "<stdout>");
		h_out = true;
		
	return 0;
}

/* vim:set expandtab cindent tabstop=4 softtabstop=4 shiftwidth=4 textwidth=0: */
